#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2015-2016 California Institute of Technology.
# Copyright (c) 2016-2020 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/klepto/blob/master/LICENSE

from test_basic import _cleanup
_cleanup()
from test_hdf import _cleanup as _cleanme
_cleanme()
from test_frame import _cleanup as _cleanframe
_cleanframe()
